// Author: Matthias Schneebeli
//////////////////////////////////////////////////////////////////////////
//
//  ROMERomeDAQ
//
//  Interface to the output of ROME.
//
//  $Id: ROMERomeDAQ.cpp 2379 2008-02-12 20:32:19Z sawada $
//
//////////////////////////////////////////////////////////////////////////

#include <RConfig.h>
#if defined( R__VISUAL_CPLUSPLUS )
#   pragma warning( push )
#   pragma warning( disable : 4800)
#endif // R__VISUAL_CPLUSPLUS
#include <TSystem.h>
#include <TROOT.h>
#include <TBranchElement.h>
#include <TMath.h>
#if defined( R__VISUAL_CPLUSPLUS )
#   pragma warning( pop )
#endif // R__VISUAL_CPLUSPLUS
#include "ROMEAnalyzer.h"
#include "ROMERomeDAQ.h"
#include "ROMETree.h"
#include "ROMETreeInfo.h"

ClassImp(ROMERomeDAQ)

//______________________________________________________________________________
ROMERomeDAQ::ROMERomeDAQ()
:ROMEDAQSystem()
,fRootFiles(0)
,fROMETrees(0)
,fTreeIndex(0)
,fInputFileNameIndex(-1)
,fCurrentTreeName("")
,fMaxEventNumber(0)
,fTreeInfo(new ROMETreeInfo())
,fTreePositionLookup(0)
,fTreeNEntries(0)
,fTimeStamp(0)
{
}

//______________________________________________________________________________
ROMERomeDAQ::~ROMERomeDAQ() {
   SafeDelete(fTreeInfo);
   SafeDeleteArray(fTreePositionLookup);
   SafeDeleteArray(fTreeNEntries);
   Int_t j;
   const Int_t nTree = gROME->GetTreeObjectEntries();
   if (fROMETrees) {
      for (j = 0; j < nTree; j++) {
         SafeDelete(fROMETrees[j]);
      }
      SafeDeleteArray(fROMETrees);
   }
}

//______________________________________________________________________________
Bool_t ROMERomeDAQ::Init() {
   if (gROME->isOnline()) {
      ROMEPrint::Error("Rome mode is not supported for online analysis.\n");
      return false;
   }
   if (gROME->isOffline()) {
      int i;
      ROMEString filename;
      const Int_t nInputFile = gROME->GetNumberOfInputFileNames();
      ROMEPrint::Print("Program is running offline.\n");
      if ((gROME->IsFileNameBasedIO() || gROME->IsRunNumberAndFileNameBasedIO())) {
         fRootFiles = new TFile*[nInputFile];
         for (i = 0; i < nInputFile; i++) {
            gROME->ConstructFilePath(gROME->GetRawInputDirString(), gROME->GetInputFileNameAt(i), filename);
            fRootFiles[i] = new TFile(filename.Data(), "READ");
            gROOT->cd();
            if (fRootFiles[i]->IsZombie()) {
               ROMEPrint::Warning("Inputfile '%s' not found.\n", filename.Data());
               return false;
            }
         }
      }
   }

   const Int_t nTree = gROME->GetTreeObjectEntries();
   fTreePositionLookup = new Long64_t*[nTree];
   fTreeNEntries = new Long64_t[nTree];

   Int_t j;
   fROMETrees = new ROMETree*[nTree];
   for (j = 0; j < nTree; j++) {
      fROMETrees[j] = new ROMETree(new TTree());
   }
   return true;
}

//______________________________________________________________________________
Bool_t ROMERomeDAQ::BeginOfRun() {
   const Int_t nTree = gROME->GetTreeObjectEntries();
   const Int_t nInputFile = gROME->GetNumberOfInputFileNames();
   Int_t run = static_cast<Int_t>(gROME->GetCurrentRunNumber());

   int i, j, k;
   for (j = 0; j < nTree; j++) {
      fROMETrees[j]->SetFileName(gROME->GetTreeObjectAt(j)->GetFileName());
      fROMETrees[j]->SetConfigInputFileName(gROME->GetTreeObjectAt(j)->GetConfigInputFileName());
      fROMETrees[j]->SetRead(gROME->GetTreeObjectAt(j)->isRead());
      fROMETrees[j]->SetName(gROME->GetTreeObjectAt(j)->GetName());
      fROMETrees[j]->AllocateBranchActive(gROME->GetTreeObjectAt(j)->GetNBranchActive());
      memcpy(fROMETrees[j]->GetBranchActive(),
             gROME->GetTreeObjectAt(j)->GetBranchActive(),
             gROME->GetTreeObjectAt(j)->GetNBranchActive() * sizeof(Bool_t));
   }

   Int_t nKey;
   if (gROME->isOffline()) {
      this->SetRunning();
      // Read Trees
      ROMEString filename;
      ROMEString treename;
      if (gROME->IsRunNumberBasedIO()) {
         fRootFiles = new TFile*[nTree];
      }
      TTree *tree;
      ROMETree *romeTree;
      ROMEString runNumberString;
      if ((gROME->IsRunNumberBasedIO() || gROME->IsRunNumberAndFileNameBasedIO()))
         gROME->GetCurrentRunNumberString(runNumberString);
      bool treeRead = false;
      for (j = 0; j < nTree; j++) {
         romeTree = fROMETrees[j];
         tree = romeTree->GetTree();
         if (romeTree->isRead()) {
            treeRead = true;

            if (gROME->IsRunNumberBasedIO()) {
               if(romeTree->GetConfigInputFileName().Length()) {
                  gROME->ConstructFilePath(gROME->GetRawInputDirString(), romeTree->GetConfigInputFileName(),
                                           filename);
               } else {
                  ROMEString tmpfilename;
                  tmpfilename.SetFormatted("%s%s.root", romeTree->GetName(), runNumberString.Data());
                  gROME->ConstructFilePath(gROME->GetRawInputDirString(), tmpfilename,
                                           filename);
               }
               gSystem->ExpandPathName(filename);
               fRootFiles[j] = new TFile(filename.Data(),"READ");
               if (fRootFiles[j]->IsZombie()) {
                  ROMEPrint::Warning("Inputfile '%s' not found.\n", filename.Data());
                  return false;
               }
               ROMEPrint::Print("Reading %s\n", filename.Data());
               if (!tree->Read(romeTree->GetName())) {
                  return false;
               }
               romeTree->SetFile(fRootFiles[j]);
               gROOT->cd();
            } else if (gROME->IsFileNameBasedIO()) {
               if (fTreeIndex > 0) {
                  if (fTreeIndex >= fRootFiles[fInputFileNameIndex]->GetNkeys()) {
                     fTreeIndex = 0;
                  } else {
                     treename.SetFormatted("%s_%d",fCurrentTreeName.Data(),fTreeIndex);
                     fRootFiles[fInputFileNameIndex]->cd();
                     if (!tree->Read(treename.Data())) {
                        return false;
                     }
                     romeTree->SetFile(fRootFiles[fInputFileNameIndex]);
                     fTreeIndex++;
                     tree->SetName(fCurrentTreeName.Data());
                  }
               }
               if (fTreeIndex == 0) {
                  fInputFileNameIndex++;
                  if (nInputFile <= fInputFileNameIndex) {
                     this->SetTerminate();
                     return true;
                  }
                  gROME->SetCurrentInputFileName(gROME->GetInputFileNameAt(fInputFileNameIndex).Data());
                  fRootFiles[fInputFileNameIndex]->cd();
                  ROMEPrint::Print("Reading %s\n",
                                   gROME->ConstructFilePath(gROME->GetRawInputDirString(),
                                                            gROME->GetCurrentInputFileName(),
                                                            filename).Data());
                  if (!tree->Read(romeTree->GetName())) {
                     return false;
                  }
                  romeTree->SetFile(fRootFiles[fInputFileNameIndex]);
                  fCurrentTreeName = romeTree->GetName();
                  fTreeIndex++;
               }
            } else if (gROME->IsRunNumberAndFileNameBasedIO()) {
               fInputFileNameIndex = -1;
               fCurrentTreeName = romeTree->GetName();
               for (i = 0; i < nInputFile; i++) {
                  nKey = fRootFiles[i]->GetNkeys();
                  for (k = 0; k < nKey; k++) {
                     treename.SetFormatted("%s",fCurrentTreeName.Data());
                     if (k > 0) {
                        treename.AppendFormatted("_%d",k);
                     }
                     fRootFiles[i]->cd();
                     if (!tree->Read(treename.Data())) {
                        return false;
                     }
                     romeTree->SetFile(fRootFiles[i]);
                     static_cast<TBranchElement*>(tree->FindBranch("Info"))->SetAddress(&fTreeInfo);
                     tree->GetBranch("Info")->GetEntry(0);
                     if (fTreeInfo->GetRunNumber() == gROME->GetCurrentRunNumber()) {
                        gROME->SetCurrentInputFileName(gROME->GetInputFileNameAt(i));
                        ROMEPrint::Print("Reading %s\n",
                                         gROME->ConstructFilePath(gROME->GetRawInputDirString(),
                                                                  gROME->GetCurrentInputFileName(),
                                                                  filename).Data());
                        tree->SetName(fCurrentTreeName.Data());
                        fInputFileNameIndex = i;
                        break;
                     }
                  }
               }
               if (fInputFileNameIndex == -1) {
                  ROMEPrint::Warning("Run "R_LLD" not found in the specified input files !\n",
                                     gROME->GetCurrentRunNumber());
                  ROMEPrint::Warning("Skipping run "R_LLD".\n",gROME->GetCurrentRunNumber());
                  this->SetEndOfRun();
                  return true;
               }
            }
            romeTree->SetTree(tree);
            fTreeNEntries[j] = tree->GetEntries();
         } else {
            fRootFiles[j] = 0;
         }
      }
      if (!treeRead) {
         ROMEPrint::Error("No input tree specified for running in rome mode.\n");
         return false;
      }
      this->ConnectTrees();

      // Get run number
      if ((gROME->IsFileNameBasedIO())) {
         for (j = 0; j < nTree; j++) {
            romeTree = fROMETrees[j];
            tree = romeTree->GetTree();
            if (romeTree->isRead()) {
               if (tree->GetEntriesFast()>0) {
                  tree->GetBranch("Info")->GetEntry(0);
                  gROME->SetCurrentRunNumber(fTreeInfo->GetRunNumber());
                  run = static_cast<Int_t>(gROME->GetCurrentRunNumber());
               }
            }
         }
      }

      // Prepare event number array
      Int_t iEvent;
      fMaxEventNumber = 0;
      fTreeInfo->SetEventNumber(0);

      // Check max event number
      for (j = 0; j < nTree; j++) {
         romeTree = fROMETrees[j];
         tree = romeTree->GetTree();
         if (romeTree->isRead()) {
            for(iEvent = 0; iEvent < fTreeNEntries[j]; iEvent++) {
               tree->GetBranch("Info")->GetEntry(iEvent);
            }
            if (fMaxEventNumber < fTreeInfo->GetEventNumber()) {
               fMaxEventNumber = fTreeInfo->GetEventNumber();
            }
         }
      }
      for (j = 0; j < nTree; j++) {
         romeTree = fROMETrees[j];
         tree = romeTree->GetTree();
         if (romeTree->isRead()) {
            fTreePositionLookup[j] = new Long64_t[static_cast<int>(fMaxEventNumber+ 1)]; // Warning : potential loss of data (int)
            for (iEvent = 0; iEvent < fMaxEventNumber + 1; iEvent++) {
               fTreePositionLookup[j][iEvent] = -1;
            }
            for(iEvent = 0; iEvent < fTreeNEntries[j]; iEvent++) {
               tree->GetBranch("Info")->GetEntry(iEvent);
               if (fTreeInfo->GetRunNumber() == run) {
                  fTreePositionLookup[j][fTreeInfo->GetEventNumber()] = iEvent;
               }
            }
         } else {
            fTreePositionLookup[j] = 0;
         }
      }
   }
   return true;
}

//______________________________________________________________________________
Bool_t ROMERomeDAQ::Event(Long64_t event) {
   if (gROME->isOffline()) {
      int j;
      ROMETree *romeTree;
      TTree *tree;
      bool found = false;
      const Int_t nTree = gROME->GetTreeObjectEntries();
      int treePosition;

      if (event > fMaxEventNumber) {
         this->SetEndOfRun();
         return true;
      }

      // read event
      for (j=0;j<nTree;j++) {
         romeTree = fROMETrees[j];
         tree = romeTree->GetTree();
         if (romeTree->isRead()) {
            treePosition = static_cast<Int_t>(fTreePositionLookup[j][event]);
            if (treePosition >= 0 &&
                treePosition < fTreeNEntries[j]) {
               found = true;
               if (gROME->IsRunNumberBasedIO()) {
                  tree->SetDirectory(fRootFiles[j]);
               } else if (gROME->IsFileNameBasedIO() || gROME->IsRunNumberAndFileNameBasedIO()) {
                  tree->SetDirectory(fRootFiles[fInputFileNameIndex]);
               }
               tree->GetEntry(treePosition);
               gROME->SetCurrentEventNumber(fTreeInfo->GetEventNumber());
            }
         }
      }

      if (!found) {
         this->SetContinue();
         return true;
      }
   }

   return true;
}

//______________________________________________________________________________
Long64_t ROMERomeDAQ::Seek(Long64_t event)
{
   if (gROME->isOffline()) {
      if (event < 0 || event > fMaxEventNumber) {
         Warning("Seek", "Event number "R_LLD" was not found.", event);
         event = gROME->GetCurrentEventNumber();
         gROME->SetDontReadNextEvent();
         SetContinue();
      }
   }
   return event;
}

//______________________________________________________________________________
Bool_t ROMERomeDAQ::EndOfRun() {
   if (gROME->isOffline()) {
      const Int_t nTree = gROME->GetTreeObjectEntries();
      int j;
      if (gROME->IsRunNumberBasedIO()) {
         for (j=0;j<nTree;j++) {
            if (fROMETrees[j]->isRead()) {
               fROMETrees[j]->GetTree()->Delete("");
               fRootFiles[j]->Close();
               fROMETrees[j]->SetTree(new TTree());
            }
            SafeDelete(fRootFiles[j]);
         }
         SafeDeleteArray(fRootFiles);
      }
      // delete sequential number array
      for (j=0;j<nTree;j++) {
         SafeDeleteArray(fTreePositionLookup[j]);
      }
   }
   return true;
}

//______________________________________________________________________________
Bool_t ROMERomeDAQ::Terminate() {
   if (gROME->isOffline()) {
   }
   return true;
}
